@@ -11,9 +11,9 @@ import datetime
 from functools import lru_cache
 
 from .exceptions import ElementPathTypeError
+from .datatypes import Timezone
 from .xpath_nodes import AttributeNode, TextNode, TypedAttribute, TypedElement, \
-    etree_iter_nodes, is_etree_element, is_element_node, is_document_node, \
-    is_attribute_node
+    etree_iter_nodes, is_etree_element, is_element_node, is_document_node
 
 
 class XPathContext(object):
@@ -21,11 +21,15 @@ class XPathContext(object):
     The XPath dynamic context. The static context is provided by the parser.
 
     Usually the dynamic context instances are created providing only the root element.
-    Variables argument is needed if the XPath expression refers to predefined variables.
+    Variable values argument is needed if the XPath expression refers to in-scope variables.
     The other optional arguments are needed only if a specific position on the context is
     required, but have to be used with the knowledge of what is their meaning.
 
     :param root: the root of the XML document, can be a ElementTree instance or an Element.
+    :param namespaces: a dictionary with mapping from namespace prefixes into URIs, \
+    used when namespace information is not available within document and element nodes. \
+    This can be useful when the dynamic context has additional namespaces and root \
+    is an Element or an ElementTree instance of the standard library.
     :param item: the context item. A `None` value means that the context is positioned on \
     the document node.
     :param position: the current position of the node within the input sequence.
@@ -35,52 +39,71 @@ class XPathContext(object):
     :param current_dt: current dateTime of the implementation, including explicit timezone.
     :param timezone: implicit timezone to be used when a date, time, or dateTime value does \
     not have a timezone.
+    :param documents: available documents. This is a dictionary from absolute URIs \
+    onto document nodes. Used by the function fn:doc.
+    :param collections: available collections. This is a dictionary from absolute \
+    URIs onto sequences of nodes. Used by the function fn:collection.
+    :param default_collection: this is the sequence of nodes used when fn:collection \
+    is called with no arguments.
     """
     _iter_nodes = staticmethod(etree_iter_nodes)
+    _parent_map = None
+    _elem = None
 
-    def __init__(self, root, item=None, position=1, size=1, axis=None,
-                 variables=None, current_dt=None, timezone=None,
-                 documents=None, collections=None, default_collection=None):
-        if not is_element_node(root) and not is_document_node(root):
-            raise ElementPathTypeError(
-                "invalid argument root={!r}, an Element is required.".format(root)
-            )
-
+    def __init__(self, root, namespaces=None, item=None, position=1, size=1, axis=None,
+                 variables=None, current_dt=None, timezone=None, documents=None,
+                 collections=None, default_collection=None):
         self.root = root
-        if item is not None:
-            self.item = item
+        self.namespaces = namespaces
+
+        if is_etree_element(root) and not callable(root.tag):
+            if item is None:
+                self.item = self._elem = root
+            elif is_etree_element(item) and not callable(root.tag):
+                self.item = self._elem = item
+            else:
+                self.item, self._elem = item, root
+        elif is_document_node(root):
+            self.item = None if item is None else item
         else:
-            self.item = root if hasattr(root, 'tag') else None
+            msg = "invalid root {!r}, an Element or an ElementTree instance required"
+            raise ElementPathTypeError(msg.format(root))
 
         self.position = position
         self.size = size
         self.axis = axis
-        self.variables = {} if variables is None else dict(variables)
-        self.current_dt = current_dt or datetime.datetime.now()
-        self.timezone = timezone
-        self.documents = {} if documents is None else dict(documents)
-        self.collections = {} if collections is None else dict(collections)
+
+        if variables is None:
+            self.variables = {}
+        else:
+            self.variables = {k: v for k, v in variables.items()}
+
+        if timezone is None or isinstance(timezone, Timezone):
+            self.timezone = timezone
+        else:
+            self.timezone = Timezone.fromstring(timezone)
+        self.current_dt = current_dt or datetime.datetime.now(tz=self.timezone)
+
+        self.documents = documents
+        self.collections = collections
         self.default_collection = default_collection
-        self._elem = item if is_element_node(item) else root
-        self._parent_map = None
 
     def __repr__(self):
-        return '%s(root=%r, item=%r, position=%r, size=%r, axis=%r)' % (
-            self.__class__.__name__, self.root, self.item, self.position, self.size, self.axis
-        )
+        return '%s(root=%r, item=%r)' % (self.__class__.__name__, self.root, self.item)
 
     def __copy__(self):
         obj = type(self)(
             root=self.root,
+            namespaces=self.namespaces,
             item=self.item,
             position=self.position,
             size=self.size,
             axis=None,
-            variables=self.variables.copy(),
+            variables=self.variables,
             current_dt=self.current_dt,
             timezone=self.timezone,
-            documents=self.documents.copy(),
-            collections=self.collections.copy(),
+            documents=self.documents,
+            collections=self.collections,
             default_collection=self.default_collection,
         )
         if self.item is None:
@@ -133,7 +156,10 @@ class XPathContext(object):
         elif isinstance(item, TypedAttribute):
             path.append('@%s' % item[0][0])
             item = self._elem
-        if isinstance(item, TypedElement):
+
+        if item is None:
+            return '' if not path else path[0]
+        elif isinstance(item, TypedElement):
             item = item[0]
 
         while True:
@@ -145,7 +171,7 @@ class XPathContext(object):
 
     def is_principal_node_kind(self):
         if self.axis == 'attribute':
-            return is_attribute_node(self.item)
+            return isinstance(self.item, (AttributeNode, TypedAttribute))
         else:
             return is_element_node(self.item)
 
@@ -164,11 +190,10 @@ class XPathContext(object):
         :param results: list containing selection results.
         """
         status = self.item, self.size, self.position
-        root = self.root.getroot() if is_document_node(self.root) else self.root
 
         self.size = len(results)
         for self.position, self.item in \
-                enumerate(self._iter_nodes(root, with_attributes=True), start=1):
+                enumerate(self._iter_nodes(self.root, with_attributes=True), start=1):
             if self.item in results:
                 yield self.item
             elif isinstance(self.item, AttributeNode):
@@ -186,7 +211,7 @@ class XPathContext(object):
 
     def iter_selector(self, selector):
         """
-        Iterator for generic selector.
+        Iterator for generic selector with buffering into a list.
 
         :param selector: a selector generator function.
         """
@@ -199,6 +224,39 @@ class XPathContext(object):
 
         self.item, self.size, self.position = status
 
+    def iter_product(self, selectors, varnames=None):
+        """
+        Iterator for cartesian products of selectors.
+
+        :param selectors: a sequence of selector generator functions.
+        :param varnames: a sequence of variables for storing the generated values.
+        """
+        iterators = [x(self) for x in selectors]
+        dimension = len(iterators)
+        prod = [None] * dimension
+        max_index = dimension - 1
+
+        k = 0
+        while True:
+            try:
+                value = next(iterators[k])
+            except StopIteration:
+                if not k:
+                    return
+                iterators[k] = selectors[k](self)
+                k -= 1
+            else:
+                try:
+                    self.variables[varnames[k]] = value
+                except (TypeError, IndexError):
+                    pass
+
+                prod[k] = value
+                if k == max_index:
+                    yield tuple(prod)
+                else:
+                    k += 1
+
     ##
     # Context item iterators for axis
 
@@ -212,7 +270,14 @@ class XPathContext(object):
 
     def iter_attributes(self):
         """Iterator for 'attribute' axis and '@' shortcut."""
-        if not is_element_node(self.item):
+        if isinstance(self.item, (AttributeNode, TypedAttribute)):
+            status = self.size, self.position, self.axis
+            self.size = self.position = 1
+            self.axis = 'attribute'
+            yield self.item
+            self.size, self.position, self.axis = status
+            return
+        elif not is_element_node(self.item):
             return
 
         status = self.item, self.size, self.position, self.axis
@@ -229,13 +294,9 @@ class XPathContext(object):
 
         self.item, self.size, self.position, self.axis = status
 
-    def iter_children_or_self(self, child_axis=False):
-        """
-        Iterator for 'child' forward axis and '/' step.
-
-        :param child_axis: if set to `True` use the child axis anyway.
-        """
-        if not child_axis and self.axis is not None:
+    def iter_children_or_self(self):
+        """Iterator for 'child' forward axis and '/' step."""
+        if self.axis is not None:
             yield self.item
             return
 
@@ -257,6 +318,10 @@ class XPathContext(object):
             self.size = len(elem)
             for self.position, self.item in enumerate(elem, start=1):
                 yield self.item
+        elif is_document_node(self.item):
+            self.size = self.position = 1
+            self.item = self.item.getroot()
+            yield self.item
 
         self.item, self.size, self.position, self.axis = status
 
@@ -322,34 +387,37 @@ class XPathContext(object):
 
         self.item, self.size, self.position, self.axis = status
 
-    def iter_descendants(self, item=None, axis=None):
+    def iter_descendants(self, axis=None):
         """
         Iterator for 'descendant' and 'descendant-or-self' forward axes and '//' shortcut.
 
-        :param item: use another item instead of the context's item.
         :param axis: the context axis, for default has no explicit axis.
         """
         status = self.item, self.size, self.position, self.axis
         self.axis = axis
+        with_self = axis != 'descendant'
 
-        if item is not None:
-            self.item = item[0] if isinstance(item, TypedElement) else item
-        elif isinstance(self.item, TypedElement):
+        if isinstance(self.item, TypedElement):
             self.item = self.item[0]
 
         if self.item is None:
-            self.size = self.position = 1
-            yield self.root
-            self.item = self.root.getroot() if is_document_node(self.root) else self.root
-        elif not is_element_node(self.item):
+            if is_document_node(self.root):
+                descendants = [x for x in self._iter_nodes(self.root, with_root=with_self)]
+                self.size = len(descendants)
+            else:
+                descendants = [x for x in self._iter_nodes(self.root)]
+                self.size = len(descendants)
+                if with_self:
+                    # Yields twice root element to emulate document node
+                    # FIXME substituting the root element with ElementTree(root)??
+                    self.position = 1
+                    yield self.root
+        elif is_element_node(self.item) or is_document_node(self.item):
+            descendants = [x for x in self._iter_nodes(self.item, with_root=with_self)]
+            self.size = len(descendants)
+        else:
             return
 
-        if axis == 'descendant':
-            descendants = [x for x in self._iter_nodes(self.item, with_root=False)]
-        else:
-            descendants = [x for x in self._iter_nodes(self.item)]
-
-        self.size = len(descendants)
         for self.position, self.item in enumerate(descendants, start=1):
             yield self.item
 
@@ -357,17 +425,17 @@ class XPathContext(object):
 
     def iter_ancestors(self, axis=None):
         """
-        Iterator for 'ancestor-or-self' and 'ancestor' reverse axes.
+        Iterator for 'ancestor' and 'ancestor-or-self' reverse axes.
 
-        :param axis: the context axis, default is 'ancestor-or-self'.
+        :param axis: the context axis, default is 'ancestor'.
         """
         status = self.item, self.size, self.position, self.axis
-        self.axis = axis or 'ancestor-or-self'
+        self.axis = axis or 'ancestor'
 
         if isinstance(self.item, TypedElement):
             self.item = self.item[0]
 
-        ancestors = [self.item] if axis == 'ancestor-or-self' else []
+        ancestors = [self.item] if self.axis == 'ancestor-or-self' else []
         parent = self.get_parent(self.item)
         while parent is not None:
             ancestors.append(parent)
